class Aufgabe21
{
  /// Die Hauptfunktion, die den ganzen beschriebenen
  /// Kodierungsprozess durchfhrt.
  //
  // Funktionsweise (Aufgabe 21):
  //
  // Ein 1D-Signal wird komplex WT-transformiert.  Siehe
  // Signal1D.ComplexWaveletTransform.  Statt der kompl. Werte x(t)
  // selbst soll (fr jedes Subband) der komplexe Quotient y(t) aus
  // zwei aufeinanderfolgenden Werten kodiert werden.  Also y(t) so,
  // dass x(t) = x(t-1) * y(t).  y(t) soll koordinatenweise
  // quantisiert und beide Koordinaten zu einem einzelnen Symbol
  // zusammengefasst werden.  Die Gre von y(t) soll nach oben
  // beschrnkt sein (vor allem wegen dem Fall x(t-1)=0).  In diesem
  // Fall soll ein Escape-Symbol geschrieben werden und (mit zweitem
  // Kontext) x(t) direkt quantisiert und kodiert werden.
  //
  // Es sind also 4 Schritte notwendig:
  //
  // 1. Komplexe WT-Transformation: Dazu wird auf
  //    ComplexWaveletTransform zurckgegriffen.
  //
  // 2. Komplexe Quotienten von aufeinanderfolgenden Werten berechnen:
  //    Wird im Moment nicht fr jedes Subband gemacht, sonder ber das
  //    ganze von ComplexWaveletTransform zurckgegebene komplexe Signal.
  //
  // 3. Kontextbildung, Koordinatenweise Quantisierung, Zusammenfassen
  //    beider Koordinaten zu einem Symbol und Escape-Symbol.
  //
  // 4. Kodierung.
  //
  // S:       Das zu kodierende Signal (z.B. Audio-Signal).
  // depth:   Die verwendete Tiefe bei der Wavelet Transformation.
  // max:     Maximum fr den Betrag des reellen oder imaginren Anteil
  //          im Quotient.
  // delta:   Quantisierungs-Delta.
  // BS:      Der Bitstream, in den geschrieben wird.

  public static void Encode (Signal1D S, int depth, double max, double delta,
    Bitstream BS)
    throws java.io.IOException
  {
    // Schritt 1: Komplexe Wavelet-Transformation
    Signal1D I = S.ComplexWaveletTransform (depth);
    
    System.out.println ("Komplexe Wavelet Transformation:");
    S.Print ();
    I.Print ();
    
    // S enthlt jetzt die reellen, I die imaginren Anteile. Die
    // fertige ComplexWaveletTransform sollte zustzlich entweder
    // eine Sammlung der verschiedener Subbands zurckgeben, oder
    // die einzelnen Subbands knnten sonst an dieser Stelle aus S
    // und I gewonnen werden. Im Moment werden einfach S und I
    // weiterverwendet.

    System.out.println ("Erhaltene Symbole:");

    Quantized1D Q = new Quantized1D (2);
    Signal1D V = new Signal1D (2);

    HuffmanCoder HC = new HuffmanCoder (BS);

    // Maximal erlaubter Betrag fr Quotienten.
    V.Set (0, max);
    V.Set (1, 0);
    Q.Quantize (V, delta);
    int mag = Q.Get (0) + 1;
    // Escape-Symbol.
    int escape = mag * mag * 4;

    double y_real = 0;
    double y_imag = 0;

    for (int i = 0; i < S.Length (); i ++)
    {
      // Komplexer Quotient.
      double x_real = S.Get (i);
      double x_imag = I.Get (i);
      
      double d = y_real * y_real + y_imag * y_imag;
      double z_real = (x_real * y_real + x_imag * y_imag) / d;
      
      double z_imag = (x_imag * y_real - x_real * y_imag) / d;

      int context = 0;

      // Kontextbildung.
      if (Math.abs (z_real) > max || Math.abs (z_imag) > max ||
          Double.isNaN(z_real) || Double.isNaN(z_imag))
      {
        context = 1;
        z_real = x_real;
        z_imag = x_imag;
      }

      System.out.println (x_real + "," + x_imag + " Q = " + z_real + "," + z_imag + " -> ");

      // Koordinatenweise quantisieren.
      V.Set (0, z_real);
      V.Set (1, z_imag);
      Q.Quantize (V, delta);
      int q_real = Q.Get (0);
      int q_imag = Q.Get (1);
      Q.Dequantize (V, delta);
      double d_real = V.Get (0);
      double d_imag = V.Get (1);

      System.out.print ("    " + q_real + "," + q_imag + " -> ");

      if (context == 0)
      {
        // Symbole zusammenfassen.
        int merged = (q_real + mag) * 2 * mag + q_imag + mag;

        // Kodieren.
        HC.Encode (merged, 0);

        double a = y_real;
        double b = y_imag;
        y_real = a * d_real - b * d_imag;
        y_imag = a * d_imag + b * d_real;

        System.out.println (merged + "(0)");
      }
      else
      {
        // Escape-Symbol fr Kontext-Switch.
        HC.Encode (escape, 0);
        System.out.print (escape + "(0), ");

        // Kodieren.
        int x = (q_real < 0 ? 1 : 0) + 2 * Math.abs (q_real);
        int y = (q_imag < 0 ? 1 : 0) + 2 * Math.abs (q_imag);
        HC.Encode (x, 1);
        HC.Encode (y, 1);
        y_real = d_real;
        y_imag = d_imag;
        System.out.println (x + "(1), " + y + "(1)");
      }
    }

    // Ende-Symbol.
    HC.Encode (escape, 0);
    HC.Encode (1, 1);
    HC.Encode (1, 1);

    HC.Close ();

  }

  /// Dekodiert ein mit Encode() kodiertes Signal wieder.
  //
  //
  // depth:   Die verwendete Tiefe bei der Wavelet Transformation.
  // max:     Maximum fr den Betrag des reellen oder imaginren Anteil
  //          im Quotient.
  // delta:   Quantisierungs-Delta.
  // BS:      Der Bitstream, aus dem gelesen wird.
  // Rckgabewert: Das wiederhergestellte Signal.

  public static Signal1D Decode (int depth, double max, double delta,
    Bitstream BS)
    throws java.io.IOException
  {
    Quantized1D Q = new Quantized1D (2);
    Signal1D V = new Signal1D (2);

    HuffmanCoder HC = new HuffmanCoder (BS);

    // Maximal erlaubter Betrag fr Quotienten
    V.Set (0, max);
    V.Set (1, 0);
    Q.Quantize (V, delta);
    int mag = Q.Get (0) + 1;
    // Escape-Symbol
    int escape = mag * mag * 4;

    double y_real = 0;
    double y_imag = 0;
       
    int i = 0;
    int len = 64;
    Signal1D S = new Signal1D (len);
    Signal1D I = new Signal1D (len);
    for (;;i++)
    {
      int symbol = HC.Decode (0);

      System.out.print (symbol + " -> ");

      if (symbol == -1)
        break;

      // Escape Symbol? Dann Kontext-Switch.
      if (symbol == escape)
      {
        // Lies zwei Symbole aus Kontext 1, dann automatisch wieder
        // Kontext1.
        int x2 = HC.Decode (1);
        int y2 = HC.Decode (1);
        int x, y;

        // Ende-Kennzeichnung?
        if (x2 == 1 & y2 == 1)
          break;

        if ((x2 & 1) == 1)
          x = - (x2 - 1) / 2;
        else
          x = x2 / 2;
        if ((y2 & 1) == 1)
          y = - (y2 - 1) / 2;
        else
          y = y2 / 2;

        System.out.println (x + "," + y + " (C1)");
  
        // Dequantisieren.
        Q.Set (0, x);
        Q.Set (1, y);
        Q.Dequantize (V, delta);
        y_real = V.Get (0);
        y_imag = V.Get (1);
      }
      else
      {
        // Symbol fr Kontext 0.
        int x = symbol / (2 * mag) - mag;
        int y = symbol % (2 * mag) - mag;

        System.out.println (x + "," + y + " (C0)");
        
        // Dequantisieren.
        Q.Set (0, x);
        Q.Set (1, y);
        Q.Dequantize (V, delta);
        double d_real = V.Get (0);
        double d_imag = V.Get (1);

        double a = y_real;
        double b = y_imag;
        y_real = a * d_real - b * d_imag;
        y_imag = a * d_imag + b * d_real;
      }

      // Vergrere Signale falls ntig.
      if (i >= len)
      {
        len += 64;
        Signal1D S_temp = new Signal1D (len);
        Signal1D I_temp = new Signal1D (len);
        S_temp.CopyFrom (S);
        I_temp.CopyFrom (I);
        S = S_temp;
        I = I_temp;
      }
      S.Set (i, y_real);
      I.Set (i, y_imag);
    }

    // Richtige Gre.
    {
      Signal1D S_temp = new Signal1D (i);
      Signal1D I_temp = new Signal1D (i);
      S_temp.CopyFrom (S);
      I_temp.CopyFrom (I);
      S = S_temp;
      I = I_temp;
    }

    System.out.println ("fertig");
  
    HC.Close ();

    System.out.println ("Rekonstruierte Signale: ");
    S.Print ();
    I.Print ();

    // Als letztes, Schritt 1 rckgngig:
    // Ursprnglich Kodiertes Signal durch umgekehrte WT wiederherstellen.
    S.ComplexWaveletReconstruction (I, depth);

    return S;
  }

  /// Eine main Methode, zum testen.
  public static void main (String[] args)
    throws java.io.IOException
  {
    Signal1D S = new Signal1D (6);
    Bitstream BS = new Bitstream ("aufgabe21.bs");
   
    S.Set (0, 1.5);
    S.Set (1, 2.5);
    S.Set (2, 3.5);
    S.Set (3, 3.5);
    S.Set (4, 2.5);
    S.Set (5, 1.5);
    S.Print ();

    System.out.println ("Kodieren:");
    
    Encode (S, 2, 10.0, 0.125, BS);

    BS.Close ();

    System.out.println ("Dekodieren:");

    BS = new Bitstream ("aufgabe21.bs");

    S = Decode (2, 10.0, 0.125, BS);

    BS.Close ();

    S.Print ();
    
  }
}

