/*
 * Clemens Wunderl
 * modified by Rade Kutil
 */


import java.io.*;

/**
 * Instances of this class support both bitwise reading and writing to a
 * random access file.
 *
 * @author <a href="mailto:cwunderl@cosy.sbg.ac.at">Clemens Wunderl</a>
 * @see    RandomAccessFile
 */
public class Bitstream {
    /*
     * As RandomAccessFile supports only bytewise reading and writing, the
     * bitstream has an internal buffer which holds the current byte. A cursor
     * keeps track of the current position of the file pointer within the
     * current byte.
     *
     * On each read (write) access, the buffer's bit which the cursor points
     * to is read (written), and the cursor is subsequently incremented by 1.
     *
     * If the cursor's value is 8 (after accessing the last bit of the current
     * byte), the buffer's content is written to the file and the cursor is
     * reset to 0.
     *
     * If the cursor's value is 0 (after flushing the buffer's content or at
     * the beginning of the file), the file's next byte is read and stored in
     * the buffer.
     *
     * Additionally, if the Bitstream is reset or closed, the buffer's content
     * is also written to the file. Note that if a 'partial' byte is written
     * (i.e. the cursor's value does not equal 8 at the time of flushing), the
     * byte is padded with zeros since files must contain whole bytes.
     */

    /**
     * A buffer holding the current byte of the file.
     */
    private int buffer = 0;

    /**
     * The current position of the file pointer relative to the current byte
     * of the file.
     */
    private byte bitCount = 0;

    /**
     * The <code>RandomAccessFile</code> representing the connection to the
     * file.
     */
    private RandomAccessFile file;

    boolean Modified;

    /**
     * Creates a bitstream for bitwise reading from and writing to a file with
     * the specified name. A new {@link RandomAccessFile} object is created to
     * represent the connection to the file. If the file does not already
     * exist then an attempt will be made to create it.
     *
     * @param     filename               the system-dependent filename
     * @exception FileNotFoundException  if the file exists but is a directory
     *                                   rather than a regular file, or cannot
     *                                   be opened or created for any other
     *                                   reason.
     */
    public Bitstream(String filename) throws FileNotFoundException
    {
        file = new RandomAccessFile(filename, "rw");
    } // Bitstream

    /**
     * Cleans up the connection to the file, and ensures that the
     * <code>close</code> method of this bitstream is called when there are no
     * more references to this stream.
     */
    protected void finalize() {
        try {
            Close();
        } catch (IOException e) {
            System.err.println("IOException when finalizing object: "
                               + e.toString());
        }
    } // finalize

    /**
     * Closes this bitstream and releases any system resources associated with
     * the stream. A closed bitstream cannot perform input or output
     * operations and cannot be reopened.
     *
     * @exception IOException  if an I/O error occurs.
     */
    public void Close() throws IOException
    {
        flush();
        file.close();
        file = null;
    } // close

    /**
     * Sets the file pointer to the beginning of the file.
     *
     * @exception IOException  if an I/O error occurs.
     */
    public void Reset() throws IOException
    {
        flush();
        file.seek(0);
    } // reset

    /**
     * Writes the contents of the buffer to the file if the current byte was
     * accessed at least once.
     *
     * @exception IOException  if an I/O error occurs.
     */
    private void flush() throws IOException
    {
        if (Modified && bitCount > 0) {
            file.write(buffer);
            buffer = 0;
            bitCount = 0;
            Modified = false;
        }
    } // flush

    /**
     * Reads a bit of data from the file at the current position of the file
     * pointer. The bit is returned as a boolean value.
     *
     * @return                  the next bit of this file as a
     *                          <code>boolean</code>.
     * @exception EOFException  if the file has reached the end.
     * @exception IOException   if an I/O error occurs.
     */
    public boolean ReadBit() throws EOFException, IOException
    {
        if (file == null) {
            throw new IOException("File is closed");
        }

        if (bitCount == 8) bitCount = 0;

        if (bitCount == 0)
        {
            buffer = file.read();
            if (buffer == -1) throw new EOFException();
        }

        return ((buffer & (1 << bitCount++)) == 0) ? false: true;
    } // readBit

    /**
     * Writes a bit to the file at the current position of the file pointer.
     *
     * @param     bit          a bit value to be written
     * @exception IOException  if an I/O error occurs.
     */
    public void WriteBit(boolean bit) throws IOException
    {
        if (file == null) {
            throw new IOException("File is closed");
        }

        if (bitCount == 8) flush();

        buffer |= (bit ? 1 : 0) << bitCount++;

        Modified = true;
    } // writeBit

    /**
     * Reads a signed 32-bit integer from the file. This method reads 32 bits
     * from the file, starting at the current file pointer.
     *
     * @return                 the next 32 bits of the file, interpreted as an
     *                         <code>int</code>.
     * @exception IOException  if an I/O error occurs.
     */
    public int ReadInt() throws IOException
    {
        int number = 0;

        for (int i = 31; i >= 0; --i) {
            number <<= 1;

            if (ReadBit()) {
                number |= 1;
            }
        }

        return number;
    } //ReadInt

    /**
     * Writes an <code>int</code> to the file as 32 bits, highest bit
     * first. The write starts at the current position of the file pointer.
     *
     * @param number           an <code>int</code> to be written.
     * @exception IOException  if an I/O error occurs.
     */
    public void WriteInt(int number) throws IOException
    {
        for (int i = 31; i >= 0; --i) {
            WriteBit(((number >> i) & 1) == 1);
        }
    } // WriteInt

    public void WriteBits (int bits, int bitCount) throws IOException
    {
        for (int i = bitCount - 1; i >= 0; --i) {
            WriteBit(((bits >> i) & 1) == 1);
        }
    }

    public int ReadBits (int bitCount) throws IOException
    {
        int bits = 0;
        for (int i = bitCount - 1; i >= 0; --i)
        {
            bits <<= 1;
            if (ReadBit ()) bits |= 1;
        }
        return bits;
    }

    /**
     * For testing purposes. Writes three <code>int</code> values to a given
     * file, resets the file, and reads three <code>int</code> values.
     *
     * @param args  a filename to write to and read from
     */
    public static void main(String[] args) {
        Bitstream bs;

        try {
            bs = new Bitstream(args[0]);
            bs.WriteInt(1100);
            bs.WriteInt(0);
            bs.WriteInt(-23);
            //bs.Reset();
            bs.Close ();
            bs = new Bitstream(args[0]);
            System.out.println(bs.ReadInt());
            System.out.println(bs.ReadInt());
            System.out.println(bs.ReadInt());
            bs.Close();
        } catch (ArrayIndexOutOfBoundsException e) {
            System.out.println("No file given:");
            e.printStackTrace();
        } catch (FileNotFoundException e) {
            System.out.println("File not found:");
            e.printStackTrace();
        } catch (EOFException e) {
            System.out.println("EOFException:");
            e.printStackTrace();
        } catch (IOException e) {
            System.out.println("IOException:");
            e.printStackTrace();
        }
    } // main

} // Bitstream
