import java.util.Vector;

class HuffmanTable
{
  Code Codes[];
  int MaxLength;

  HuffmanTable ()
  {
  }


  /**
   * Constructor: Builds up the huffman table
   * @param symStat SymbolStatistics
   * @param context The context to code
   * @author Markus Watzl
   */
  HuffmanTable (SymbolStatistics symStat, int context) {
    int symRg = symStat.GetSymbolRange(context);
    float symCt = symStat.GetSymbolCount(context);
    Codes = new Code[symRg];
    SymbolBlock tempSBlk;
    Symbol tempSymb;
    Vector symBlks = new Vector();
    double minPr = 1.0, min2Pr = 1.0, temp;
    int min = 0, min2 = 1;

    // Put each symbol with its frequency in a seperate SymbolBlock
    for (int aSym = 0; aSym < symRg; aSym++) {
      if (symStat.GetSymbolCount(aSym, context) > 0) {
        tempSBlk = new SymbolBlock(((float) symStat.GetSymbolCount(aSym, context)) / symCt);
        tempSBlk.symbols.addElement(new Symbol(aSym));
        symBlks.add(tempSBlk);
      }
    }

    // Loop until only 1 SymbolBlock is left:
    while (symBlks.size() > 1)
    {
      minPr = 1.0;
      min2Pr = 1.0;
      // Search for the 2 smallest percentages
      for (int i = 0; i < symBlks.size(); i++) {
        temp = ((SymbolBlock) symBlks.get(i)).prct;
        if (temp < minPr) {
          min2Pr = minPr;
          minPr = temp;
          min2 = min;
          min = i;
        } else if (temp < min2Pr) {
          min2Pr = temp;
          min2 = i;
        }
      }

      // Append 0 to smallest and 1 to second smallest SymbolBlock
      ((SymbolBlock) symBlks.get(min2)).addCode(true);
      ((SymbolBlock) symBlks.get(min)).addCode(false);

      // Combine smallest and second smallest SymbolBlock
      ((SymbolBlock) symBlks.get(min2)).add((SymbolBlock) symBlks.get(min));
      symBlks.remove(min);
    }

    // Finally: Write codes into the "Codes"-Array and define MaxLength
    for (int i = 0; i < symBlks.size(); i++) {
      tempSBlk = (SymbolBlock) symBlks.get(i);
      for (int j = 0; j < tempSBlk.symbols.size(); j++) {
        tempSymb = (Symbol) tempSBlk.symbols.get(j);
        if (MaxLength < tempSymb.codeLength)
          MaxLength = tempSymb.codeLength;
        Codes[tempSymb.index] = new Code(tempSymb.code, tempSymb.codeLength);
      }
    }
  }


  /**
   * Inner class: Includes code, codelength and index of one Symbol
   * @author Markus Watzl
   */
  private class Symbol {
    public int index = 0;
    public int code = 0;
    public int codeLength = 0;

    /**
     * Creates new Symbol
     * @param ind Index of the new Symbol
     */
    Symbol(int ind) {
      index = ind;
    }

    /**
     * Appends one bit at the beginning of the code
     * @param b The bit to be appended
     */
    public void appendCode(boolean b) {
      codeLength++;
      if (b) {
        code+= (1 << (codeLength - 1));
      }
    }
  }


  /**
   * Inner class: A block of Symbols of undefined length with combined frequency
   * @author Markus Watzl
   */
  private class SymbolBlock {
    public double prct;
    public Vector symbols = new Vector();
    SymbolBlock(double pct) {
      prct = pct;
    }

    /**
     * Appends one SymbolBlock
     * @param sB SymbolBlock to append
     */
    public void add(SymbolBlock sB) {
      prct+= sB.prct;
      for (int i = 0; i < sB.symbols.size(); i++) {
        symbols.add(sB.symbols.get(i));
      }
    }

    /**
     * Appends a bit to all Symbols in this block
     * @param cod Bit to append (true = 1, false = 0)
     */
    public void addCode(boolean cod) {
      for (int i = 0; i < symbols.size(); i++) {
        ((Symbol) symbols.get(i)).appendCode(cod);
      }
    }
  }


  Code GetCode (int Symbol)
  {
    return Codes[Symbol];
  }

  int FindCode (Code C)
  {
    if (MaxLength < C.GetBitLength ())
       return -1;
    for (int i = 0; i < Codes.length; i ++)
        if (Codes[i] != null && Codes[i].Equals (C))
           return i;
    return -1;
  }

  void Encode (Bitstream BS) throws java.io.IOException
  {
    int lengthBits = 1;
    for (int l = 1; l < MaxLength; l *= 2)
        lengthBits ++;

    BS.WriteInt (Codes.length);
    BS.WriteInt (MaxLength);
    if (MaxLength > 0)
       for (int i = 0; i < Codes.length; i ++)
       {   int l = 0;
           if (Codes[i] != null)
              l = Codes[i].GetBitLength ();
           BS.WriteBits (l, lengthBits);
           if (l > 0) Codes[i].Encode (BS);
       }
  }

  void Decode (Bitstream BS) throws java.io.IOException
  {
    int codeCount = BS.ReadInt ();
    Codes = new Code[codeCount];
    MaxLength = BS.ReadInt ();

    int lengthBits = 1;
    for (int l = 1; l < MaxLength; l *= 2)
        lengthBits ++;
    
    if (MaxLength > 0)
       for (int i = 0; i < codeCount; i ++)
       {   int l = BS.ReadBits (lengthBits);
           if (l > 0)
           {  Codes[i] = new Code ();
              for (int j = 0; j < l; j ++)
                  Codes[i].ReadBit (BS);
           }
       }
    else // there is only one symbol => symbol length = 0
       Codes[codeCount-1] = new Code ();
  }

  void Print ()
  {
    for (int i = 0; i < Codes.length; i ++)
        if (Codes[i] != null)
        {  System.out.print (i);
           System.out.print (": ");
           GetCode (i).Print ();
           System.out.println ();
        }
  }

  public static void main (String[] args)
  {
    SymbolStatistics SS = new SymbolStatistics ();
    SS.AddSymbol (0, 0);
    SS.AddSymbol (2, 0);
    SS.AddSymbol (5, 0);
    SS.AddSymbol (5, 0);
    SS.AddSymbol (7, 0);
    SS.AddSymbol (7, 0);
    SS.AddSymbol (7, 0);
    SS.AddSymbol (9, 0);
    SS.Print ();
    HuffmanTable HT = new HuffmanTable (SS, 0);
    HT.Print ();
  }
}
