/* Image Predictor fuer zwei verschiedene Modi *?

/* Modi sind:
   1 ...         a a a
                 a b         mit vier Koeffizienten

   2 ...           a
                 a a a
               a a b         mit sechs Koeffizienten
*/

/* Anmerkungen:
   1) Konstruktor: Angabe des Modus und des zu kodierenden Bilds

   2) Coder: Erzeugung eines Symbol-Streams aus dem Originalbild
     * Input: -
     * Output: das kodierte Bild als Quantized2D

   3) Decoder: Erzeugung des Originalbilds aus dem Symbol-Stream
     * Input: -
     * Output: ein Signal2D (Bild)
     
   4) zur Prediction:
     * es werden Samples von allen Bildpunkten des Originalbilds
       an einen OptimalPredictor uebergeben
     * dieser berechnet die besten Koeffizienten
     * diese Koeffizienten werden quantisiert
     * mit den quantisierten Koeffizienten wird das Bild kodiert
       (und auch wieder dekodiert)
*/

/* Zeilen die mit //*d beginnen, beinhalten Debug Anweisungen. */


class ImagePredictor
{
    // Modus: 1 oder 2
    private int mode;  
    // Originalbild - wird gendert!
    private Signal2D input;
    // Output
    private Quantized2D output;
    // "persoenlicher Prediktor"
    private OptimalPredictor optPred;
    // Koeffizienten (quantisiert)
    private Quantized1D optCoeffs;
    // Quantisierungsfaktor
    private double q;
    private double delta;
    
    
    // Konstruktor
    ImagePredictor(Signal2D image, int mode, double delta) 
    {
        // Originalbild setzen
        this.input = image;
        // Quantisierungsfaktor setzen
        if (delta != 0) this.delta = delta;
        else this.delta = 0.5;
        this.q = 1.0 / this.delta;
        // Modus: 1 oder 2
        // und dabei den OptimalPredictor und die optCoeffs initialisieren
        if (mode >= 2)
        {
            this.mode = 2;
            optPred = new OptimalPredictor(6);
            optCoeffs = new Quantized1D(6);
        }
        else
        {
            this.mode = 1;
            optPred = new OptimalPredictor(4);
            optCoeffs = new Quantized1D(4);
        }
        // output alloziieren
        output = new Quantized2D(input.Dimension());
    }


    // kodiert das Originalbild unter Beruecksichtigung des Modus
    // und gibt die entstehende Symbolfolge zurueck
    Quantized2D Code()
    {
        // optimale quantisierte Koeffizienten holen (nach optCoeffs)
        GetOptimalCoeffs();
        // ... und diese wieder im OptimalPredictor speichern
        for (int i = 0; i < optCoeffs.Length(); i++)
        {
            optPred.Set(i, (double) optCoeffs.Get(i));
        }
        //*d*/ System.out.print("coeffs of predictor after quantization: "); optPred.Print();
        //*d*/ System.out.println("");

        Position pixel;
        Signal1D vector;
        double predictedValue;
        int quantizedValue;
        // jetzt das Bild predicten!
        for (int row = 0; row < input.Dimension().Y(); row++)
        {
            for (int col = 0; col < input.Dimension().X(); col++)
            {
                pixel = new Position(col, row); 
                // hier die Differenz (quant. Originalwert - Prediction) abspeichern
                vector = GetSurroundingVector(input, pixel);
                predictedValue = optPred.Predict(vector);
                //*d*/ System.out.print("code - vector: "); vector.Print();
                //*d*/ System.out.print("code - predicted: "); 
                //*d*/ System.out.println(predictedValue); 
                quantizedValue = Quant(input.Get(pixel) - predictedValue);
                output.Set(pixel, quantizedValue);
                // dequantisierten Wert + Voraussage gleich wieder ins input speichern
                // (wiederhergestellte Werte)
                input.Set(pixel, Dequant(quantizedValue) + predictedValue);
            }
        }
        
        //*d*/ System.out.println("changed input image:"); input.Print();
        //*d*/ System.out.println(""); 
        return output;
    }


    // dekodiert das vorher kodierte Bild wieder und retourniert
    // ein Bild das sich im Inhalt wahrscheinlich vom Originalbild 
    // unterscheidet...
    Signal2D Decode()
    {
        // Rueckgabebild mit denselben Dimensionen wie das Originalbild
        Signal2D ret = new Signal2D(input.Dimension());
        
        Position pixel;
        double predictedValue;
        // jetzt: rueck-predicten (mit den quantisierten Koeffizienten optCoeffs)
        for (int row = 0; row < input.Dimension().Y(); row++)
        {
            for (int col = 0; col < input.Dimension().X(); col++)
            {
                pixel = new Position(col, row);
                predictedValue = optPred.Predict(GetSurroundingVector(ret, pixel));
                //*d*/ System.out.print("decode - predicted: ");
                //*d*/ System.out.println(predictedValue); 
                ret.Set(pixel, Dequant(output.Get(pixel)) + predictedValue);
            }
        }        
        
        return ret;
        
    } 


    // gibt zu einem bestimmten Pixel in einem Bild, b, die Koeffizienten a
    // in einem Signal1D zurueck. Anzahl der Koeffizienten haengt vom
    // Modus ab.
    private Signal1D GetSurroundingVector(Signal2D image, Position pixel)
    {
        Signal1D ret;
        Position pos;
        if (mode == 1) {
            // Schema: a a a
            //         a b
            ret = new Signal1D(4);
            // Werte einfuellen: falls ausserhalb des Bildes, 0
            for (int i = 0; i <4; i++)
            {
                // Position der Koeffizienten
                if (i == 0) pos = new Position(pixel.X()-1, pixel.Y()-1);
                else if (i == 1) pos = new Position(pixel.X(), pixel.Y()-1);
                else if (i == 2) pos = new Position(pixel.X()+1, pixel.Y()-1);
                else pos = new Position(pixel.X()-1, pixel.Y());
                // Werte korrigieren und einfuellen
                if ( (pos.X() < 0) || (pos.X() >= image.Dimension().X()) || 
                     (pos.Y() < 0) || (pos.Y() >= image.Dimension().Y()) ) 
                    ret.Set(i, 0);
                else ret.Set(i, image.Get(pos));
            }
        }
        else // mode == 2
        {
            // Schema:     a
            //           a a a
            //         a a b
            ret = new Signal1D(6);
            // Werte einfuellen: falls ausserhalb des Bildes, 0
            for (int i = 0; i <6; i++)
            {
                // Position der Koeffizienten
                if (i == 0) pos = new Position(pixel.X(), pixel.Y()-2);
                else if (i == 1) pos = new Position(pixel.X()-1, pixel.Y()-1);
                else if (i == 2) pos = new Position(pixel.X(), pixel.Y()-1);
                else if (i == 3) pos = new Position(pixel.X()+1, pixel.Y()-1);
                else if (i == 4) pos = new Position(pixel.X()-2, pixel.Y());
                else pos = new Position(pixel.X()-1, pixel.Y());
                // Werte korrigieren und einfuellen
                if ( (pos.X() < 0) || (pos.X() >= image.Dimension().X()) || 
                     (pos.Y() < 0) || (pos.Y() >= image.Dimension().Y()) ) 
                    ret.Set(i, 0);
                else ret.Set(i, image.Get(pos));
            }
        }
        return ret;
    }


    // Koeffizienten vom OptimalPredictor berechnen lassen
    // und in optCoeffs ablegen    
    private void GetOptimalCoeffs()
    {
        // durch alle Pixel durchgehen und jeweils das Trainingset
        // an den Optimal Predictor uebertragen
        Position p;
        for (int col = 0; col < input.Dimension().X(); col++)
        {
            for (int row = 0; row < input.Dimension().Y(); row++)
            {
                p = new Position(col,row);
                optPred.AddTrainingSet( GetSurroundingVector(input, p),
                                        input.Get(p));
            }
        }
        // optimale Koeffizienten berechnen
        optPred.CalcCoeffs();
        //*d*/ System.out.println(""); System.out.print("optimal coeffs: "); optPred.Print();
        // Koeffizienten quantisieren
        optCoeffs.Quantize(optPred, delta);
        //*d*/ System.out.print("quantized coeffs: "); optCoeffs.Print(); System.out.println("");
    }


    // Quantisieren eines Double-Werts
    private int Quant(double val)
    {
       	return (int) Math.round(val * q);
    }


    // Dequantisieren eines Integer-Werts
    private double Dequant(int val)
    {
        return ((double) val) * delta;
    }
    
    // DIE main Prozedur zum Testen der Funktionalitaet
    public static void main (String[] args)
    {
        // Testbild erstellen
        // ist 3x4 Matrix mit:   12.4   14.2   18.0   20.9
        //                       14.3   17.1   19.9   22.0
        //                       16.2   19.2   23.9   25.1
        Signal2D testImage = new Signal2D(new Position(4,3));
        testImage.Set(new Position(0,0), 12.4);
        testImage.Set(new Position(0,1), 14.3);
        testImage.Set(new Position(0,2), 16.2);
        testImage.Set(new Position(1,0), 14.2);
        testImage.Set(new Position(1,1), 17.1);
        testImage.Set(new Position(1,2), 19.2);
        testImage.Set(new Position(2,0), 18.0);
        testImage.Set(new Position(2,1), 19.9);
        testImage.Set(new Position(2,2), 23.9);
        testImage.Set(new Position(3,0), 20.9);
        testImage.Set(new Position(3,1), 22.0);
        testImage.Set(new Position(3,2), 25.1);
        
        System.out.println("image to transform:");
        testImage.Print();

        // ImagePredictor erstellen: Modus 2, Quantisierungsfaktor 0.25
        ImagePredictor predictor = new ImagePredictor(testImage, 2, 0.25);
        
        // nun predicten
        Quantized2D result = predictor.Code();
        System.out.println("");
        System.out.println("quantized prediction result:"); 
        result.Print();
        
        System.out.println("");
        // rcktransformieren
        Signal2D retransformedImage;
        retransformedImage = predictor.Decode();
        System.out.println("");
        System.out.println("re-predicted image:");
        retransformedImage.Print();
        System.out.println("");
    }

}
