/**
  Encodes a Int2DArray completely (no quantisation errors)
**/
class IntQuadTreeCoder
{
  private Int2DArray Image;
  private int Max;
  private Quantized2D Status;
  private EntropyCoder Coder;

  final static int INSIGNIFICANT = 0;
  final static int SIGNIFICANT = 1;
  final static int PARTIAL = 2;

  final static int NODE_CONTEXT = 0;
  final static int LEAF_CONTEXT = 1;
  final static int REFINEMENT_CONTEXT = 2;
  final static int SIGN_CONTEXT = 3;

  IntQuadTreeCoder (Int2DArray Image)
  {
    this.Image = Image;
  }

  void Encode (EntropyCoder Coder)
  {
    this.Coder = Coder;

    // Max must be smallest 2-power *greater* than maximum of image
    int IMax = Image.AbsMax ();
    Max = 1;
    while (Max <= IMax) Max *= 2;

    Coder.WriteInt ((int) Max);
    Status = new Quantized2D (Image.Dimension ());
    int threshold = Max;
    while (threshold > 1)
    {
      threshold /= 2;
      EncodeBitplane (new Int2DSubArray (Image, new Position (0, 0), Image.Dimension ()), threshold, PARTIAL);
    }
  }

  void EncodeBitplane (Int2DSubArray A, int threshold, int ParentStatus)
  {
    //A.GetOrigin ().Print ();
    //A.Dimension ().Print ();
    //System.out.println (ParentStatus);

    int NewStatus = ParentStatus;

    if (ParentStatus == PARTIAL)
    {
      int NodeContext = NODE_CONTEXT;
      if (A.Size () == 1)
         NodeContext = LEAF_CONTEXT;

      int SigCount = 0;
      int InsigCount = 0;
      for (PositionIterator i = new PositionIterator (A.Dimension ()); ! i.AtEnd (); i.Iterate ())
      {
        if (Status.Get (A.GetOrigin ().Plus (i.Current ())) != SIGNIFICANT)
        {  if   (Math.abs (A.Get (i.Current ())) >= threshold)
                SigCount ++;
           else InsigCount ++;
        }
      }

      if (SigCount > 0)
      {  if (InsigCount > 0)
         {  Coder.Encode (PARTIAL, NodeContext);
            NewStatus = PARTIAL;
         }
         else
         {  Coder.Encode (SIGNIFICANT, NodeContext);
            NewStatus = SIGNIFICANT;
         }
      }
      else
      {  if (InsigCount > 0)
         {  Coder.Encode (INSIGNIFICANT, NodeContext);
            NewStatus = INSIGNIFICANT;
         }
         else
         {  NewStatus = INSIGNIFICANT; // just not PARTIAL
         }
      }
    }

    if (A.Size () > 1)
    {
       for (int i = 0; i < 4; i ++)
           EncodeBitplane (A.Quadrant (i), threshold, NewStatus);
    }
    else
    {
       Position p = A.GetOrigin ();
       if (Status.Get (p) == INSIGNIFICANT)
       {  if (NewStatus == SIGNIFICANT)
          {  Coder.Encode (Image.Get (p) > 0 ? 1 : 0, SIGN_CONTEXT);
             Status.Set (p, NewStatus);
          }
       }
       else
       {  Coder.Encode ((Image.Get (p) & threshold) == 0 ? 0 : 1, REFINEMENT_CONTEXT);
       }
    }
  }

  void Decode (EntropyCoder Coder)
    throws java.io.IOException
  {
    this.Coder = Coder;
    Max = Coder.ReadInt ();
    Status = new Quantized2D (Image.Dimension ());
    int threshold = Max;
    while (threshold > 1)
    {
      threshold /= 2;
      DecodeBitplane (new Int2DSubArray (Image, new Position (0, 0), Image.Dimension ()), threshold, PARTIAL);
    }
  }

  void DecodeBitplane (Int2DSubArray A, int threshold, int ParentStatus)
    throws java.io.IOException
  {
    //A.GetOrigin ().Print ();
    //A.Dimension ().Print ();
    //System.out.println (ParentStatus);

    int NewStatus = ParentStatus;

    if (ParentStatus == PARTIAL)
    {
      int NodeContext = NODE_CONTEXT;
      if (A.Size () == 1)
         NodeContext = LEAF_CONTEXT;

      int InsigCount = 0;
      for (PositionIterator i = new PositionIterator (A.Dimension ()); ! i.AtEnd (); i.Iterate ())
        if (Status.Get (A.GetOrigin ().Plus (i.Current ())) == INSIGNIFICANT)
           InsigCount ++;

      if (InsigCount > 0)
         NewStatus = Coder.Decode (NodeContext);
    }

    if (A.Size () > 1)
    {
       for (int i = 0; i < 4; i ++)
           DecodeBitplane (A.Quadrant (i), threshold, NewStatus);
    }
    else
    {
       Position p = A.GetOrigin ();
       if (Status.Get (p) == INSIGNIFICANT)
       {  if (NewStatus == SIGNIFICANT)
          {  if   (Coder.Decode (SIGN_CONTEXT) == 1)
                  Image.Set (p, +threshold);
             else Image.Set (p, -threshold);
             Status.Set (p, NewStatus);
          }
       }
       else
       {  
          int val = Image.Get (p);
          if (Coder.Decode (REFINEMENT_CONTEXT) != 0)
             Image.Set (p, val + (val > 0 ? threshold : -threshold));
       }
    }
  }
  

  static void main (String[] args)
    throws java.io.IOException
  {
    Quantized2D S = new Quantized2D (new Position (2, 2));
    S.Set (new Position (0, 0), 15);
    S.Set (new Position (1, 0),  4);
    S.Set (new Position (0, 1),  7);
    S.Set (new Position (1, 1),  1);

    Bitstream B = new Bitstream ("hugh");
    HuffmanCoder C = new HuffmanCoder (B);
    IntQuadTreeCoder Q = new IntQuadTreeCoder (S);
    Q.Encode (C);
    C.Close ();

    //DebugCoder DC = new DebugCoder ();
    //IntQuadTreeCoder DQ = new IntQuadTreeCoder (S);
    //DQ.Encode (DC);

    B.Reset ();
    S.SetAll (0);

    IntQuadTreeCoder Q2 = new IntQuadTreeCoder (S);
    Q2.Decode (C);
    S.Print ();
    System.out.println ();
  }
}
