public class VectorQuantizer {
    
    private int dim; // Dimension of vectors

    private int bpv; // bits per vector

    private double[][] codebook; // the quantization vectors

    private double dist; // average distortion


    public VectorQuantizer (int Dimension, int BitsPerVector)
    {
      bpv = BitsPerVector;
      dim = Dimension;
    }

    public int GetBitsPerVector ()
    {
      return bpv;
    }
    
    public int GetDimension ()
    {
      return dim;
    }

    public double GetDistortion ()
    {
      return dist;
    }
    
    public double[] GetCodeBookVector (int Index)
    {
      return codebook[Index];
    }

    // calculates optimal quantization vectors with LBG Algorithm
    public void LBG (Signal1D[] S)
    {
        int numberTrainingVec = S[0].Length ();
        if (dim != S.length)
           throw new IllegalArgumentException("have to specify one signal for each dimension");
        
        int codebookSize = 1 << bpv;
        //System.out.println("codebookSize is " + codebookSize +".");
        
        codebook = new double[codebookSize][dim];
        int[] bin = new int[codebookSize];
        double[][] yy = new double[codebookSize][dim];

        if (numberTrainingVec < codebookSize )
           throw new IllegalArgumentException("number of training vecors smaller than codebook size");

        int mtemp=1; // Anzahl der momentan vorh. Vektoren
        for (int k=0; k < dim; ++k)
        {
          codebook[0][k]=0.0;
          for (int i=0; i < numberTrainingVec; ++i)
              codebook[0][k] += S[k].Get (i);
          codebook[0][k] /= (double) numberTrainingVec;
        }

        double totd1 = 0.0;
        while (mtemp < codebookSize)
        {
          mtemp *= 2; // mtemp verdoppeln

          // splitting
          for (int i=0; i < mtemp; ++i)
          {
            int j = i / 2;
            double delta = 0.001;
            if ((i & 1) == 1) { delta = -0.001; }
            for (int k=0; k < dim; ++k)
              yy[i][k] = codebook[j][k] * (1.0 + delta);
          }

          double totd2 = 10.0e15;

          boolean flag = false;
          while (!flag)
          {
            totd1=0.0;
            for (int i=0; i < mtemp; ++i) {
              for (int k=0; k < dim; ++k)
                codebook[i][k]=0.0;
              bin[i]=0;
            }
            for (int i=0; i < numberTrainingVec; ++i){
              int j = 0;
              double dist2 = 0.0;
              for (int k=0;k < dim; ++k) {
                double c = S[k].Get(i) - yy[j][k];
                dist2 += c * c;
              }
              int index = 0;
              for (j=1; j < mtemp; ++j){
                double dist1 = 0.0;
                for (int k=0; k < dim; ++k) {
                  double c = S[k].Get(i) - yy[j][k];
                  dist1 += c * c;
                }
                if(dist1 < dist2){
                  dist2 = dist1;
                  index = j;
                }
              }
              ++bin[index];
              for (int k=0; k < dim; ++k)
                codebook[index][k]+= S[k].Get (i);
              totd1 = totd1 + dist2;
            }  // end for i<numberTrainingVec

            for (int j=0; j < mtemp; ++j){
              if(bin[j] > 0) {
                for (int k=0; k < dim; ++k){
                  codebook[j][k] /= (float) bin[j];
                  yy[j][k] = codebook[j][k];
                }
              }
            }

            totd1 /= numberTrainingVec*dim;
            double drel = (totd2 - totd1)/totd1;

            //System.out.println("codebookSize=" + mtemp + " Avg. Dist = " 
            //    + totd1 + ";  SNR = " + (-10.0*Math.log(totd1)/Math.log(10)) + ";  dre1 = " + drel + ".");
            
            flag = true;
            if (drel > 0.001) {
                flag = false;
                totd2 = totd1;
            }
        
          } /* end while(!flag) */

        } /* end while mtemp<codebookSize */

        dist = totd1;
    }
    
    
    public Quantized1D Quantize(Signal1D[] S)
    {
      Quantized1D Q = new Quantized1D (S[0].Length ());
      
      double summe = 0;
      double minSumme = Double.POSITIVE_INFINITY;
      double[] tempCbVector = null;
      int cbIndex = -1; // Index des CB-Vektors, der am besten approx.

      for (int i = 0; i < S[0].Length(); i++) { // alle Vektoren durchlaufen
          // fuer jeden dieser Vektoren den Codebook-Vektor (bzw dessen Index)
          // finden, der minimalen Abstand zum Vektor hat:
          
          // einen Vektor nehmen und Abstand zu allen Codebook-Vekt. berechnen:
          for (int j = 0; j < codebook.length; j++) { // laufe alle CB-Vektoren durch
            tempCbVector = codebook[j];
            
            // Laufe die Dimensionen durch und berechne Abstand!
            for (int k = 0; k < dim; k++)
            {   double c = S[k].Get(i) - tempCbVector[k];
                summe += c * c; // quadr.
            }
            //System.out.println("Vergl. mit CB-Vektor #" + j + ", Abstand = " + summe);
            
            // falls summe < minSumme, dann haben ich CB-Vektor gefunden,
            // der geg. Vektor besser approx.:
            if (summe < minSumme)
            {
                minSumme = summe;
                cbIndex = j; // merke mir Index des besten CB-Vektors.
            }
            
            summe = 0; // fuer naechsten CB-Vektor-Vergleich
          }
      
        // Index des am besten approx. CB-Vektors in Q eintragen.
        Q.Set(i,cbIndex);
        // fuer naechsen Input-Vektor wieder resetten:
        minSumme = Double.POSITIVE_INFINITY;
      }
      return Q;
    }
    
    public void Dequantize(Quantized1D Q, Signal1D[] S)
    {
      for (int i = 0; i < Q.Length (); i ++)
          for (int d = 0; d < dim; d ++)
              S[d].Set (i, codebook[Q.Get (i)][d]);
    }
    
    // start encodevectors
    public void EncodeVectors(Bitstream BS) {
        // codebook[][] in BS schreiben
        try {
            BS.WriteInt(this.dim);
            BS.WriteInt(this.bpv);
        } catch (Exception e) {
            System.out.println(e);
        }
        for (int i = 0; i < codebook.length; i++) {
            for (int j = 0; j < dim; j++) {
                writeDoubleToBitstream(codebook[i][j], BS);
            }
        }


    }

    private void writeDoubleToBitstream(double vektor, Bitstream BS) {
        long myVektor;
        myVektor = Double.doubleToLongBits(vektor);


        try {
            // schreibe myVektor (long(64 Bit)) als 2 Integer(32 Bit)
            BS.WriteInt((int) (myVektor & 0xFFFFFFFF));
            BS.WriteInt((int) ((myVektor >> 32) & 0xFFFFFFFF));

        } catch (Exception e) {
            System.out.println(e);
        }
    }

    public void DecodeVectors(Bitstream BS) {
        // codebook[][] aus BS lesen
        try {
            dim = BS.ReadInt ();
            bpv = BS.ReadInt ();
            codebook = new double[1 << bpv][dim];
        } catch (Exception e) {
            System.out.println(e);
        }
        for (int i = 0; i < codebook.length; i++) {
            for (int j = 0; j < dim; j++) {
                codebook[i][j] = readDoubleFromBitstream(BS);
            }
        }
    }

    private double readDoubleFromBitstream(Bitstream BS) {
        //long myVektor = BS.ReadInt() + BS.ReadInt();
        long myVektor = 0;
        try {
            myVektor = ((long) BS.ReadInt() + (((long) BS.ReadInt()) << 32));
        } catch (Exception e) {
            System.out.println(e);
        }
        return (Double.longBitsToDouble(myVektor));
    }

    // end encodevectors

    public void Print ()
    {
      for (int i = 0; i < codebook.length; i ++)
      {
        System.out.print (i +": (");
        for (int d = 0; d < dim; d ++)
        {   if (d > 0) System.out.print (", ");
            System.out.print (codebook[i][d]);
        }
        System.out.println (")");
      }
    }

    static void main (String[] args)
    {
      Signal1D[] S = new Signal1D[2];
      S[0] = new Signal1D (5);
      S[1] = new Signal1D (5);
      for (int i = 0; i < 5; i++)
      {   S[0].Set (i, i);
          S[1].Set (i, 6 - i);
      }

      System.out.println ("Originalsignal (2-D Vektoren):");
      S[0].Print ();
      S[1].Print ();

      VectorQuantizer VQ = new VectorQuantizer (2, 2);
      VQ.LBG (S);

      System.out.println ("\nCodebook:");
      VQ.Print ();

      Quantized1D Q = VQ.Quantize (S);

      System.out.println ("\nQuantisiert:");
      Q.Print ();

      VQ.Dequantize (Q, S);

      System.out.println ("\nDequantisiert:");
      S[0].Print ();
      S[1].Print ();
    }
}
